/**
 * @ecma 8+
 * @version 1.0.4
 * @author J. Taniguchi
 * @license CC https://creativecommons.org/licenses/by/4.0/
 */
class jDatePicker
{
    /**
     * 
     * @param {Object} options Initialization options.
     * @returns {jDataPicker} Instance of.
     */
    constructor(options)
    {
        /** @var {Object} OPTIONS Holds initialization options. */
        this.OPTIONS = this._validateOptions(options);

        this.UI = {
            iconLeft: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#666" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><polyline points="11 17 6 12 11 7"></polyline><polyline points="18 17 13 12 18 7"></polyline></svg>',
            iconRight: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#666" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><polyline points="13 17 18 12 13 7"></polyline><polyline points="6 17 11 12 6 7"></polyline></svg>',
            iconClose: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#fff" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><line x1="18" y1="6" x2="6" y2="18"></line><line x1="6" y1="6" x2="18" y2="18"></line></svg>',
            iconCheck: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#fff" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><polyline points="20 6 9 17 4 12"></polyline></svg>',
            iconPlus: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#d1d1d1" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><line x1="12" y1="5" x2="12" y2="19"></line><line x1="5" y1="12" x2="19" y2="12"></line></svg>',
            iconMinus: '<svg xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24" fill="none" stroke="#d1d1d1" stroke-width="2.5" stroke-linecap="round" stroke-linejoin="round"><line x1="5" y1="12" x2="19" y2="12"></line></svg>'
        };
    }

    /**
     * Open calendar.
     * 
     * @public
     * @returns {void}
     */
    open()
    {
        this._renderUi();
        this._renderMonth();
    }

    /**
     * Render main body structure.
     * 
     * @private
     * @returns {void}
     */
    _renderUi()
    {
        const self = this;

        // main structure.
        this.UI.wrap = document.createElement('div');
        this.UI.box = document.createElement('div');
        this.UI.calendarWrap = document.createElement('div');
        this.UI.year = {
            wrap: document.createElement('div'),
            title: document.createElement('div'),
            label: document.createElement('label'),
            previousBtn: document.createElement('button'),
            nextBtn: document.createElement('button')
        };
        this.UI.month = {
            wrap: document.createElement('div'),
            title: document.createElement('label'),
            previousBtn: document.createElement('button'),
            nextBtn: document.createElement('button')
        };
        this.UI.date = document.createElement('p');
        this.UI.clock = {
            wrap: document.createElement('div'),
            hour: document.createElement('p'),
            minute: document.createElement('p'),
            semi: document.createElement('div'),
            hourPlusBtn: document.createElement('button'),
            hourMinusBtn: document.createElement('button'),
            minutePlusBtn: document.createElement('button'),
            minuteMinusBtn: document.createElement('button')
        };
        this.UI.acceptBtn = document.createElement('button');
        this.UI.cancelBtn = document.createElement('button');

        this.UI.wrap.id = 'jdatepicker';
        this.UI.box.id = 'jdatepicker-box';
        this.UI.calendarWrap.id = 'jdatepicker-calendar__wrap';
        this.UI.year.wrap.className = 'jdatepicker-calendar--yearwrap';
        this.UI.year.title.className = 'jdatepicker-calendar--yeartitle';
        this.UI.month.wrap.className = 'jdatepicker-calendar--monthwrap';
        this.UI.date.className = 'jdatepicker-calendar--datelabel';
        this.UI.clock.wrap.className = 'jdatepicker-clock';
        this.UI.acceptBtn.id = 'jdatepicker-clock--accept';
        this.UI.acceptBtn.className = 'jdatepicker-controlbtn';
        this.UI.cancelBtn.id = 'jdatepicker-clock--cancel';
        this.UI.cancelBtn.className = 'jdatepicker-controlbtn';
        this.UI.clock.hour.id = 'jdatepicker-clock--hour';
        this.UI.clock.hour.className = 'jdatepicker-clock--p';
        this.UI.clock.minute.id = 'jdatepicker-clock--minute';
        this.UI.clock.minute.className = 'jdatepicker-clock--p';
        this.UI.clock.semi.id = 'jdatepicker-clock--semi';
        this.UI.clock.semi.className = 'jdatepicker-clock--p';

        if (this.OPTIONS.enableTimeChange) {
            this.UI.clock.hourPlusBtn.id = 'jdatepicker-clock-hplus',
            this.UI.clock.hourPlusBtn.className = 'jdatepicker-clock--btn';
            this.UI.clock.hourMinusBtn.id = 'jdatepicker-clock-hminus';
            this.UI.clock.hourMinusBtn.className = 'jdatepicker-clock--btn';
            this.UI.clock.minutePlusBtn.id = 'jdatepicker-clock-mplus';
            this.UI.clock.minutePlusBtn.className = 'jdatepicker-clock--btn';
            this.UI.clock.minuteMinusBtn.id = 'jdatepicker-clock-mminus';
            this.UI.clock.minuteMinusBtn.className = 'jdatepicker-clock--btn';
        }

        this.UI.wrap.appendChild(this.UI.box);
        this.UI.box.appendChild(this.UI.calendarWrap);
        this.UI.calendarWrap.appendChild(this.UI.year.wrap);
        this.UI.year.wrap.appendChild(this.UI.year.title);

        if (this.OPTIONS.enableYearChange) {
            this.UI.year.title.appendChild(this.UI.year.previousBtn);
        }

        this.UI.year.title.appendChild(this.UI.year.label);

        if (this.OPTIONS.enableYearChange) {
            this.UI.year.title.appendChild(this.UI.year.nextBtn);
        }

        this.UI.year.wrap.appendChild(this.UI.month.wrap);

        if (this.OPTIONS.enableMonthChange) {
            this.UI.month.wrap.appendChild(this.UI.month.previousBtn);
        }

        this.UI.month.wrap.appendChild(this.UI.month.title);

        if (this.OPTIONS.enableMonthChange) {
            this.UI.month.wrap.appendChild(this.UI.month.nextBtn);
        }

        this.UI.year.wrap.appendChild(this.UI.date);
        this.UI.box.appendChild(this.UI.clock.wrap);
        this.UI.box.appendChild(this.UI.acceptBtn);
        this.UI.box.appendChild(this.UI.cancelBtn);
        this.UI.clock.wrap.appendChild(this.UI.clock.hour);
        this.UI.clock.wrap.appendChild(this.UI.clock.minute);
        this.UI.clock.wrap.appendChild(this.UI.clock.semi);

        if (this.OPTIONS.enableTimeChange) {
            this.UI.clock.wrap.appendChild(this.UI.clock.hourPlusBtn);
            this.UI.clock.wrap.appendChild(this.UI.clock.hourMinusBtn);
            this.UI.clock.wrap.appendChild(this.UI.clock.minutePlusBtn);
            this.UI.clock.wrap.appendChild(this.UI.clock.minuteMinusBtn);
        }

        this.UI.year.label.innerHTML = this.OPTIONS.year;

        if (this.OPTIONS.enableYearChange) {
            this.UI.year.previousBtn.innerHTML = this.UI.iconLeft;
            this.UI.year.nextBtn.innerHTML = this.UI.iconRight;
        }

        this.UI.month.previousBtn.innerHTML = this.UI.iconLeft;

        if (this.OPTIONS.enableMonthChange) {
            this.UI.month.nextBtn.innerHTML = this.UI.iconRight;
            this.UI.month.title.innerHTML = this.OPTIONS.translation.month[this.OPTIONS.month - 1];
        }

        this.UI.acceptBtn.innerHTML = this.UI.iconCheck;
        this.UI.cancelBtn.innerHTML = this.UI.iconClose;
        this.UI.clock.hour.innerHTML = (this.OPTIONS.hour < 10 ? `0${this.OPTIONS.hour}` : this.OPTIONS.hour);
        this.UI.clock.minute.innerHTML = (this.OPTIONS.minute < 10 ? `0${this.OPTIONS.minute}` : this.OPTIONS.minute);
        this.UI.clock.semi.innerHTML = ':';

        if (this.OPTIONS.enableTimeChange) {
            this.UI.clock.hourPlusBtn.innerHTML = this.UI.iconPlus;
            this.UI.clock.hourMinusBtn.innerHTML = this.UI.iconMinus;
            this.UI.clock.minutePlusBtn.innerHTML = this.UI.iconPlus;
            this.UI.clock.minuteMinusBtn.innerHTML = this.UI.iconMinus;
        }

        document.body.appendChild(this.UI.wrap);

        // year control listeners.
        if (this.OPTIONS.enableYearChange) {
            this.UI.year.previousBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                let syear = self.OPTIONS.year - 1;

                if (syear < 1900) {
                    syear = 1900;
                }

                self.OPTIONS.onYear(syear);
                self.OPTIONS.year = syear;
                self._handleYearChange();
            });
            
            this.UI.year.nextBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                let syear = self.OPTIONS.year + 1;

                self.OPTIONS.onYear(syear);
                self.OPTIONS.year = syear;
                self._handleYearChange();
            });
        }

        // month control listeners.
        if (this.OPTIONS.enableMonthChange) {
            this.UI.month.previousBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                let smonth = self.OPTIONS.month - 1;
                self.OPTIONS.month = smonth;

                if (smonth < 1) {
                    smonth = 12;
                }

                self.OPTIONS.onMonth(smonth);
                self._handleMonthChange();
            });
            
            this.UI.month.nextBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                let smonth = self.OPTIONS.month + 1;
                self.OPTIONS.month = smonth;

                if (smonth > 12) {
                    smonth = 1;
                }

                self.OPTIONS.onMonth(smonth);
                self._handleMonthChange();
            });
        }

        // cancel listener.
        this.UI.cancelBtn.addEventListener('click', e => {
            e.preventDefault();
            e.stopImmediatePropagation();

            self.OPTIONS.onCancel();

            self.UI.wrap.remove();
        });

        // accept listener.
        this.UI.acceptBtn.addEventListener('click', e => {
            e.preventDefault();
            e.stopImmediatePropagation();

            self.OPTIONS.onAccept({
                date: self.OPTIONS.date,
                month: self.OPTIONS.month,
                year: self.OPTIONS.year,
                hour: self.OPTIONS.hour,
                minute: self.OPTIONS.minute
            });

            self.UI.wrap.remove();
        });

        // time control listeners.
        if (this.OPTIONS.enableTimeChange) {
            this.UI.clock.hourPlusBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                self.OPTIONS.hour += 1;

                if (self.OPTIONS.hour > 23) {
                    self.OPTIONS.hour = 0;
                }

                self.UI.clock.hour.innerHTML = (self.OPTIONS.hour < 10 ? `0${self.OPTIONS.hour}` : self.OPTIONS.hour);

                self.OPTIONS.onTime({
                    hour: self.OPTIONS.hour,
                    minute: self.OPTIONS.minute
                });
            });

            this.UI.clock.hourMinusBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                self.OPTIONS.hour -= 1;

                if (self.OPTIONS.hour < 0) {
                    self.OPTIONS.hour = 23;
                }
                
                self.UI.clock.hour.innerHTML = (self.OPTIONS.hour < 10 ? `0${self.OPTIONS.hour}` : self.OPTIONS.hour);

                self.OPTIONS.onTime({
                    hour: self.OPTIONS.hour,
                    minute: self.OPTIONS.minute
                });
            });

            this.UI.clock.minutePlusBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                self.OPTIONS.minute += 1;

                if (self.OPTIONS.minute > 59) {
                    self.OPTIONS.minute = 0;
                }

                
                self.UI.clock.minute.innerHTML = (self.OPTIONS.minute < 10 ? `0${self.OPTIONS.minute}` : self.OPTIONS.minute);

                self.OPTIONS.onTime({
                    hour: self.OPTIONS.hour,
                    minute: self.OPTIONS.minute
                });
            });
            
            this.UI.clock.minuteMinusBtn.addEventListener('click', e => {
                e.preventDefault();
                e.stopImmediatePropagation();

                self.OPTIONS.minute -= 1;

                if (self.OPTIONS.minute < 0) {
                    self.OPTIONS.minute = 59;
                }

                
                self.UI.clock.minute.innerHTML = (self.OPTIONS.minute < 10 ? `0${self.OPTIONS.minute}` : self.OPTIONS.minute);

                self.OPTIONS.onTime({
                    hour: self.OPTIONS.hour,
                    minute: self.OPTIONS.minute
                });
            });
        }
    }

    /**
     * Handle month change.
     * 
     * @private
     * @returns {void}
     */
    _handleMonthChange()
    {
        let monthIndex = this.OPTIONS.month - 1;

        if (monthIndex < 0) {
            monthIndex = 11;
            this.OPTIONS.month = monthIndex + 1;
        }

        if (monthIndex > 11) {
            monthIndex = 0;
            this.OPTIONS.month = monthIndex + 1;
        }

        let lastDayofMonth = this._getLastDayOfMonth(this.OPTIONS.year, this.OPTIONS.month);

        if (this.OPTIONS.date > lastDayofMonth) {
            const dates = this.UI.dates;

            let i, c;

            for (i=0; i < dates.length; i++) {
                c = parseInt(dates[i].getAttribute('data-date'));

                if (c === 1) {
                    dates[i].dispatchEvent(new Event('click'));
                    break;
                }
            }
        }

        this.UI.month.title.innerHTML = this.OPTIONS.translation.month[monthIndex];
        this._clearWheelDates();
        this._renderMonth();
    }

    /**
     * Handle year change.
     * 
     * @private
     * @returns {void}
     */
    _handleYearChange()
    {
        this.UI.year.label.innerHTML = this.OPTIONS.year;
        this._clearWheelDates();
        this._renderMonth();
        
        let lastDayofMonth = this._getLastDayOfMonth(this.OPTIONS.year, this.OPTIONS.month);

        if (this.OPTIONS.date > lastDayofMonth) {
            const dates = this.UI.dates;

            let i, c;

            for (i=0; i < dates.length; i++) {
                c = parseInt(dates[i].getAttribute('data-date'));

                if (c === 1) {
                    dates[i].dispatchEvent(new Event('click'));
                    break;
                }
            }
        }
    }

    /**
     * Render the month calendar on calendar's wheel based on options year and month values.
     * 
     * @private
     * @returns {void}
     */
    _renderMonth()
    {
        const lastDayofMonth = this._getLastDayOfMonth(this.OPTIONS.year, this.OPTIONS.month);
        const today = new Date();
        const self = this;

        let i, u, day, week, wrapSize, deg, origin, radius, d, addevent;

        wrapSize = 290;
        deg = 360 / lastDayofMonth;
        radius = wrapSize / 2;
        origin = 3;

        this.UI.dates = [];

        for (i=1; i <= lastDayofMonth; i++) {
            day = document.createElement('div');
            day.innerHTML = i;
            day.className = 'jdatepicker-calendar--button';
            day.style.height = `${radius}px`;
            day.style.transformOrigin = '0 100%';
            day.style.transform = `rotate(${origin}deg)`;
            day.style.left += '50%';
            day.setAttribute('data-date', i);
            day.setAttribute('data-month', this.OPTIONS.month);
            day.setAttribute('data-year', this.OPTIONS.year);

            this.UI.dates.push(day);

            if (today.getUTCDate() === i && this.OPTIONS.year === today.getUTCFullYear() && this.OPTIONS.month === today.getUTCMonth() + 1) {
                day.classList.add('today');
            }

            d = new Date(`${this.OPTIONS.month}/${i}/${this.OPTIONS.year}`);

            week = document.createElement('div');
            week.innerHTML = this.OPTIONS.translation.week[d.getDay()];
            day.appendChild(week);
            day.setAttribute('data-week', this.OPTIONS.translation.week[d.getDay()]);

            this.UI.calendarWrap.appendChild(day);

            origin += deg; 

            // set holiday.
            for (u=0; u < this.OPTIONS.holidays.length; u++) {
                if (this.OPTIONS.holidays[u].date === i && this.OPTIONS.month === this.OPTIONS.holidays[u].month) {
                    day.setAttribute('data-holidaydate', this.OPTIONS.holidays[u].date);
                    day.setAttribute('data-holidaymonth', this.OPTIONS.holidays[u].month);
                    day.setAttribute('title', this.OPTIONS.holidays[u].name);
                    day.classList.add('holiday');
                }
            }

            // disable dates: ignore click on disabled and add cursor.
            for (u=0; u < this.OPTIONS.disabledDates.length; u++) {
                addevent = true;

                if (this.OPTIONS.disabledDates[u].date === i && this.OPTIONS.disabledDates[u].month === this.OPTIONS.month && this.OPTIONS.year === this.OPTIONS.disabledDates[u].year) {
                    addevent = false;
                    break;
                }
            }
            
            if (addevent) {
                day.addEventListener('click', e => {
                    e.preventDefault();
                    e.stopImmediatePropagation();

                    let date = parseInt(e.currentTarget.getAttribute('data-date')),
                        week = e.currentTarget.getAttribute('data-week');

                    self.OPTIONS.date = date;
                    self.OPTIONS.onDate({
                        date: date,
                        week: week
                    });

                    let dayList = document.getElementsByClassName('jdatepicker-calendar--button');

                    for (u=0; u < dayList.length; u++) {
                        dayList[u].classList.remove('selected');
                    }

                    e.currentTarget.classList.add('selected');
                    self.UI.date.innerHTML = `<span>${week}</span> ${date}`;
                });
            } else {
                day.classList.add('disabled');
            }

            // mark checked date.
            if (i === this.OPTIONS.date) {
                day.classList.add('selected');
                this.UI.date.innerHTML = `<span>${week.innerHTML}</span> ${i}`;
            }
        }
    }

    /**
     * Remove all month's dates from wheel.
     * 
     * @private
     * @returns {void}
     */
    _clearWheelDates()
    {
        let i, count = this.UI.dates.length;

        for (i=0; i < count; i++) {
            this.UI.dates[i].remove();
        }

        this.UI.dates = [];
    }

    /**
     * Get the last day of the month.
     * 
     * @private
     * @param {Number} year 
     * @param {Number} month 
     * @returns {Number}
     */
    _getLastDayOfMonth(year, month)
    {
        return new Date(year, month, 0).getDate();
    }

    /**
     * Validate initialization options values.
     * 
     * @private
     * @param {Object} opt 
     * @returns {Object} 
     */
    _validateOptions(opt)
    {
        this.OPTIONS = opt;

        const curDate = new Date();
        const currentYear = curDate.getUTCFullYear();
        const currentMonth = curDate.getUTCMonth() + 1;
        
        // define initialization options.
        const options = {
            year: currentYear,
            month: currentMonth,
            date: curDate.getUTCDate(),
            hour: curDate.getUTCHours(),
            minute: curDate.getUTCMinutes(),
            disabledDates: [
                {
                    date: 25,
                    month: 9,
                    year: 2021
                },
                {
                    date: 1,
                    month: 1,
                    year: 2022
                }],
            enableYearChange: true,
            enableMonthChange: true,
            enableTimeChange: true,
            translation: {
                month: ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'Dezember'],
                week: ['sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat']
            },
            holidays: [
                {
                    date: 25,
                    month: 12,
                    name: 'Christmas'
                },
                {
                    date: 1,
                    month: 9,
                    name: 'Saint J'
                }
            ],
            onDate: () => {},
            onMonth: () => {},
            onYear: () => {},
            onTime: () => {},
            onCancel: () => {},
            onAccept: () => {}
        };

        if (typeof this.OPTIONS !== 'object') {
            this.OPTIONS = options;
            return options;
        }

        if (this.OPTIONS.year > 1900 && this.OPTIONS.year < currentYear + 50) {
            options.year = this.OPTIONS.year;
        }

        if (this.OPTIONS.month > 0 && this.OPTIONS.month < 13) {
            options.month = this.OPTIONS.month;
        }

        if (this.OPTIONS.date > 0 && this.OPTIONS.date < 32) {
            options.date = this.OPTIONS.date;
        }

        if (this.OPTIONS.hour > -1 && this.OPTIONS.hour < 25) {
            options.hour = this.OPTIONS.hour;
        }

        if (this.OPTIONS.minute > -1 && this.OPTIONS.minute < 60) {
            options.minute = this.OPTIONS.minute;
        }

        if (typeof this.OPTIONS.onDate === 'function') {
            options.onDate = this.OPTIONS.onDate;
        }

        if (typeof this.OPTIONS.onMonth === 'function') {
            options.onMonth = this.OPTIONS.onMonth;
        }

        if (typeof this.OPTIONS.onYear === 'function') {
            options.onYear = this.OPTIONS.onYear;
        }

        if (typeof this.OPTIONS.onTime === 'function') {
            options.onTime = this.OPTIONS.onTime;
        }

        if (typeof this.OPTIONS.onCancel === 'function') {
            options.onCancel = this.OPTIONS.onCancel;
        }

        if (typeof this.OPTIONS.onAccept === 'function') {
            options.onAccept = this.OPTIONS.onAccept;
        }

        if (typeof this.OPTIONS.enableYearChange === 'boolean') {
            options.enableYearChange = this.OPTIONS.enableYearChange;
        }

        if (typeof this.OPTIONS.enableMonthChange === 'boolean') {
            options.enableMonthChange = this.OPTIONS.enableMonthChange;
        }

        if (typeof this.OPTIONS.enableTimeChange === 'boolean') {
            options.enableTimeChange = this.OPTIONS.enableTimeChange;
        }

        if (typeof this.OPTIONS.translation === 'object' && typeof this.OPTIONS.translation.week === 'object' && typeof this.OPTIONS.translation.month === 'object') {
            if (this.OPTIONS.translation.week.length === 7) {
                options.translation.week = this.OPTIONS.translation.week;
            }

            if (this.OPTIONS.translation.month.length === 12) {
                options.translation.month = this.OPTIONS.translation.month;
            }
        }

        if (typeof this.OPTIONS.holidays === 'object' && this.OPTIONS.holidays.length > 0) {
            options.holidays = this.OPTIONS.holidays;
        }

        if (typeof this.OPTIONS.disabledDates === 'object' && this.OPTIONS.disabledDates.length > 0) {
            options.disabledDates = this.OPTIONS.disabledDates;
        }

        this.OPTIONS = options;

        return this.OPTIONS;
    }
}